#include "calo_orbit_processor.h"

#include <cassert>

#include "cmssw/FRDFileHeader_v2.h"

/* clang-format off */
// Goes through orbit worth of data and fills the output memory with the calo
// data corresponding to the non-empty bunch-crossings, as marked in bx_vect
CaloOrbitProcessor::FillOrbitMetadata CaloOrbitProcessor::FillOrbit(orbit_trailer *trailer,
                                                                    char *rd_ptr, char *wr_ptr,
                                                                    const char *rd_end_ptr,
                                                                    const char *wr_end_ptr) {
  std::pair<uint32_t, bool> orbit_header = std::pair<uint32_t, bool>{
      ProcessOrbitHeader(rd_ptr)};  //.second is the warning test enable bit
  rd_ptr += 32;                     // +32 to account for orbit header
  if (cmsswHeaders) {
    wr_ptr += sizeof(FRDFileHeader_v2);
  }  // reserving space for cmssw orbit header

  uint32_t orbit = uint32_t{orbit_header.first} -
                   1;  // Starting with orbit number one lower than what is in the header
                       // because the "link orbit" contains a few BXs of the previous orbit
  auto counts = uint32_t{0};
  uint32_t filled_bxs = 0;
  // We loop over the BX map from the orbit trailer and then match the filled
  // BXs to the data we got. The logic below is annoyingly convoluted: The first
  // BX we get in the data stream is from BX 3555, however the BX map starts at
  // BX 1, we therefore need to start reading the BX map from the 3555th bit.
  // 3555//32 = 111, so we start at the 111th word; 3555 mod 32 = 3, however we
  // start counting at BX1, so the start bit is 2.
  uint32_t bx = 3554;  // We start at 3554 here, because we increment by 1
                       // immediately after starting the loops.
  size_t word = 111;
  size_t bit = 1;  // Will be incremented immediately after starting the loop
  for (size_t pseudo_bx = 0; pseudo_bx < 3564;
       ++pseudo_bx) {                       // Looping over at most an entire orbit here.
    if (word == 14 * 8 - 1 && bit == 11) {  // Bit 11 in word 111 is the last valid BX (3564), so we
                                            // roll over afterwards. (== 11 because that's the one
                                            // we worked on in the previous loop iteration)
      word = 0;
      bit = 0;
      bx = 0;  // Will be immediately incremented to 1.
      ++orbit;
    } else if (bit < 31) {
      ++bit;
    } else {
      bit = 0;
      ++word;
    }
    ++bx;
    if ((trailer->bx_map[word] & (1 << bit)) == 0) {
      continue;  // If the bit is zero that BX was empty and we continue.
    }
    assert(rd_ptr + sizeof(blockCalo) - 1 <= rd_end_ptr);
    ++filled_bxs;
    auto *bl = reinterpret_cast<blockCalo *>(rd_ptr);
    if (bl->calo0[0] == constants::beefdead) {
      break;
    }  // orbit trailer has been reached, end of orbit data
    assert(wr_ptr + (3 + 8 * 7) * 4 - 1 <= wr_end_ptr);
    auto header = uint32_t{orbit_header.second};  // header can be added to later
    memcpy(wr_ptr, (char *)&header, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&bx, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&orbit, 4);
    wr_ptr += 4;
    for (uint32_t i = 0; i < 8; i++) {
      memcpy(wr_ptr, (char *)&i, 4);
      wr_ptr += 4;  // gives link number, can later be used to find object type
      memcpy(wr_ptr, (char *)&bl->calo0[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo1[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo2[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo3[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo4[i], 4);
      wr_ptr += 4;
      memcpy(wr_ptr, (char *)&bl->calo5[i], 4);
      wr_ptr += 4;
    }
    counts += 1;
    rd_ptr += sizeof(blockCalo);
  }

  return {counts, orbit, filled_bxs};
}