#include "muon_orbit_processor.h"

#include <cassert>

#include "log.h"

/* clang-format off */
// Goes through orbit worth of data and fills the output memory with the muons
// corresponding to the non-empty bunch-crossings, as marked in bx_vect
MuonOrbitProcessor::FillOrbitMetadata MuonOrbitProcessor::FillOrbit(orbit_trailer *trailer,
                                                                  char *rd_ptr, char *wr_ptr,
                                                                  const char *rd_end_ptr,
                                                                  const char *wr_end_ptr) {
  std::pair<uint32_t, bool> orbit_header = std::pair<uint32_t, bool>{
      ProcessOrbitHeader(rd_ptr)};  //.second is the warning test enable bit
  rd_ptr += 32;                     // +32 to account for orbit header
  if (cmsswHeaders) {
    wr_ptr += 28;
  }  // reserving space for cmssw orbit header

  uint32_t orbit = uint32_t{orbit_header.first} -
                   1;  // Starting with orbit number one lower than what is in the header
                       // because the "link orbit" contains a few BXs of the previous orbit
  auto counts = uint32_t{0};
  uint32_t filled_bxs = 0;
  // We loop over the BX map from the orbit trailer and then match the filled
  // BXs to the data we got. The logic below is annoyingly convoluted: The first
  // BX we get in the data stream is from BX 3555, however the BX map starts at
  // BX 1, we therefore need to start reading the BX map from the 3555th bit.
  // 3555//32 = 111, so we start at the 111th word; 3555 mod 32 = 3, however we
  // start counting at BX1, so the start bit is 2.
  uint32_t bx = 3554;  // We start at 3554 here, because we increment by 1
                       // immediately after starting the loops.
  size_t word = 111;
  size_t bit = 1;  // Will be incremented immediately after starting the loop
  for (size_t pseudo_bx = 0; pseudo_bx < 3564;
       ++pseudo_bx) {                       // Looping over at most an entire orbit here.
    if (word == 14 * 8 - 1 && bit == 11) {  // Bit 11 in word 111 is the last valid BX (3564), so we
                                            // roll over afterwards. (== 11 because that's the one
                                            // we worked on in the previous loop iteration)
      word = 0;
      bit = 0;
      bx = 0;  // Will be immediately incremented to 1.
      ++orbit;
    } else if (bit < 31) {
      ++bit;
    } else {
      bit = 0;
      ++word;
    }
    ++bx;
    if ((trailer->bx_map[word] & (1 << bit)) == 0) {
      continue;  // If the bit is zero that BX was empty and we continue.
    }
    assert(rd_ptr + sizeof(blockMuon) - 1 <= rd_end_ptr);
    ++filled_bxs;
    auto *bl = reinterpret_cast<blockMuon *>(rd_ptr);
    if (bl->orbit[0] == constants::beefdead) {
      break;
    }  // orbit trailer has been reached, end of orbit data
    assert(wr_ptr + (3 + 2 * 8 * 3) * 4 - 1 <=
           wr_end_ptr);  // Assuming the maximum size a decoded muon block can use.
    int mAcount = 0;
    int mBcount = 0;
    bool AblocksOn[8];
    bool BblocksOn[8];
    for (unsigned int i = 0; i < 8; i++) {
      auto pt = uint32_t{(bl->mu1f[i] >> shifts::pt) & masks::pt};

      AblocksOn[i] = ((pt > 0) || (doZS == 0));
      if ((pt > 0) || (doZS == 0)) {
        mAcount++;
      }
      pt = (bl->mu2f[i] >> shifts::pt) & masks::pt;
      BblocksOn[i] = ((pt > 0) || (doZS == 0));
      if ((pt > 0) || (doZS == 0)) {
        mBcount++;
      }
    }
    if (mAcount + mBcount == 0) {
      rd_ptr += sizeof(blockMuon);
      LOG(WARNING) << '#' << nbPackets
                   << ": Detected a bx with zero muons, this should not "
                      "happen. Packet is skipped.";
      continue;
    }

    // header word of packed muon data contains mAcount (number of muons in
    // words 3,4) and mBcount (number of muons in words 5,5), as well as the
    // warning test enable flag.
    auto header =
        uint32_t{(mAcount << 16) + ((static_cast<uint32_t>(orbit_header.second)) << 8) + mBcount};

    counts += mAcount;
    counts += mBcount;
    memcpy(wr_ptr, (char *)&header, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&bx, 4);
    wr_ptr += 4;
    memcpy(wr_ptr, (char *)&orbit, 4);
    wr_ptr += 4;
    for (unsigned int i = 0; i < 8; i++) {
      if (AblocksOn[i]) {
        memcpy(wr_ptr, (char *)&bl->mu1f[i], 4);
        wr_ptr += 4;
        memcpy(wr_ptr, (char *)&bl->mu1s[i], 4);
        wr_ptr += 4;
        // next creating mu.extra which is a copy of bl->bx with a change to the
        // first bit
        memcpy(wr_ptr, (char *)&(bl->bx[i] &= ~0x1), 4);
        wr_ptr += 4;  // set bit 0 to 0 for first muon
      }
    }

    for (unsigned int i = 0; i < 8; i++) {
      if (BblocksOn[i]) {
        memcpy(wr_ptr, (char *)&bl->mu2f[i], 4);
        wr_ptr += 4;
        memcpy(wr_ptr, (char *)&bl->mu2s[i], 4);
        wr_ptr += 4;
        // next creating mu.extra which is a copy of bl->bx with a change to the
        // first bit
        memcpy(wr_ptr, (char *)&(bl->bx[i] |= 0x1), 4);
        wr_ptr += 4;  // set bit 0 to 1 for second muon
      }
    }
    rd_ptr += sizeof(blockMuon);
  }

  return {counts, orbit, filled_bxs};
}